use std::{fs, io::{self, BufRead, BufReader, Write}};

fn main() {
    let args : Vec<String> = std::env::args().collect();
    if args.len() < 3 {
        println!("usage: {} <patch_file> <patch_code>", args[0]);
        return;
    }
    let codes = match read_codes(args[2].as_str()) {
        Ok(code) => code,
        Err(e) => {
            println!("read code error: {}", e.to_string());
            return;
        }
    };
    if let Err(e) = verify_code(&codes) {
        println!("verify code error: {}", e.to_string());
        return;
    }
    if let Err(e) = apply_code(args[1].as_str(), &codes) {
        println!("apply code error: {}", e.to_string());
    }
}

fn read_codes(code_file : &str) -> io::Result<Vec<(u32, u32)>> {
    let file = fs::File::open(code_file).expect("cannot open patch_code");
    let reader = BufReader::new(file);
    let mut codes = Vec::new();
    let mut lines = 1u32;
    for line in reader.lines() {
        if let Ok(l) = line {
            let r_str = l.as_str();
            let code_vec : Vec<&str> = r_str.split(' ').collect();
            if code_vec.len() == 2 {
                let val1 = u32::from_str_radix(code_vec[0], 16);
                let val2 = u32::from_str_radix(code_vec[1], 16);
                if val1.is_ok() && val2.is_ok() {
                    codes.push((lines, val1.unwrap()));
                    codes.push((lines, val2.unwrap()));
                }
            } // else skip
            lines += 1;
        }
    }
    Ok(codes)
}

fn verify_code(codes : &Vec<(u32, u32)>) -> Result<(), String> {
    let mut pos = 0usize;
    let max_len = codes.len();
    let mut search_code = false;
    while pos < max_len {
        let line = codes[pos].0;
        let first_code = codes[pos].1;
        let op_code = (first_code >> 28) as u8;
        match op_code {
            0x0 | 0x1 |0x2 => {
                if pos + 2 > max_len {
                    return Err(format!("wrong code at line {}: unexpected end", line));
                }
                pos += 2;
            }
            0x4 => {
                if pos + 4 > max_len {
                    return Err(format!("wrong code at line {}: unexpected end", line));
                }
                let next_code = codes[pos + 2].1;
                if next_code & 0xf0000000 != 0x40000000 {
                    return Err(format!("serial code error at line {}: should start with 0x4", line + 1));
                }
                pos += 4;
            }
            0x5 => {
                if pos + 4 > max_len {
                    return Err(format!("wrong code at line {}: unexpected end", line));
                }
                pos += 4;
            }
            0x8 => {
                let search_bytes = (first_code & 0xffff) as usize;
                if search_bytes == 0 {
                    return Err(format!("wrong code at line {}: search length should not be 0", line));
                }
                let mut code_len = (search_bytes - 1) / 4 + 2;
                code_len += code_len % 2;
                if pos + code_len > max_len {
                    return Err(format!("wrong code at line {}: unexpected end", line));
                }
                search_code = true;
                pos += code_len;
            }
            0xa => {
                if pos + 1 >= max_len {
                    return Err(format!("wrong code at line {}: unexpected end", line));
                }
                let write_bytes = codes[pos + 1].1 as usize;
                if write_bytes == 0 {
                    return Err(format!("wrong code at line {}: write length should not be 0", line));
                }
                if write_bytes > 320 {
                    return Err(format!("wrong code at line {}: write length should not exceed 320", line));
                }
                let code_len = ((write_bytes - 1) / 8 + 1) * 2 + 2;
                if pos + code_len > max_len {
                    return Err(format!("wrong code at line {}: unexpected end", line));
                }
                pos += code_len;
            }
            _ => {
                return Err(format!("unknown code at line {}", line));
            }
        }
        
    }
    Ok(())
}

fn apply_code(patch_file : &str, codes : &Vec<(u32, u32)>) -> Result<(), String> {
    let fn_map_err = |e : io::Error| format!("apply code error: {}", e.to_string());
    let mut data = fs::read(patch_file).map_err(fn_map_err)?;
    let mut pos = 0usize;
    let max_code_len = codes.len();
    let max_data_len = data.len();
    let mut search_addr : Option<usize> = None;
    let mut code_len = 0;
    loop {
        pos += code_len;
        if pos >= max_code_len {
            break;
        }
        let line = codes[pos].0;
        let first_code = codes[pos].1;
        let op_code = (first_code >> 28) as u8;
        match op_code {
            0x0 => {
                code_len = 2;
                let val = codes[pos + 1].1 as u8;
                let addr = if (first_code & 0xf000000) == 0x8000000 {
                    if search_addr == None {
                        continue;
                    }
                    (first_code & 0xffffff) as usize + search_addr.unwrap()
                } else {
                    (first_code & 0xfffffff) as usize
                };
                if addr >= max_data_len {
                    return Err(format!("apply code error at line {}: address overflow", line));
                }
                data[addr] = val;
            }
            0x1 => {
                code_len = 2;
                let val = codes[pos + 1].1 as u16;
                let addr = if (first_code & 0xf000000) == 0x8000000 {
                    if search_addr == None {
                        continue;
                    }
                    (first_code & 0xffffff) as usize + search_addr.unwrap()
                } else {
                    (first_code & 0xfffffff) as usize
                };
                if addr + 1 >= max_data_len {
                    return Err(format!("apply code error at line {}: address overflow", line));
                }
                data[addr] = (val & 0xff) as u8;
                data[addr + 1] = ((val >> 8) & 0xff) as u8;
            }
            0x2 => {
                code_len = 2;
                let val = codes[pos + 1].1;
                let addr = if (first_code & 0xf000000) == 0x8000000 {
                    if search_addr == None {
                        continue;
                    }
                    (first_code & 0xffffff) as usize + search_addr.unwrap()
                } else {
                    (first_code & 0xfffffff) as usize
                };
                if addr + 3 >= max_data_len {
                    return Err(format!("apply code error at line {}: address overflow", line));
                }
                data[addr] = (val & 0xff) as u8;
                data[addr + 1] = ((val >> 8) & 0xff) as u8;
                data[addr + 2] = ((val >> 16) & 0xff) as u8;
                data[addr + 3] = ((val >> 24) & 0xff) as u8;
            }
            0x4 => {
                code_len = 4;
                let mut val = codes[pos + 1].1;
                let mut addr = if (first_code & 0xf000000) >= 0x8000000 {
                    if search_addr == None {
                        continue;
                    }
                    (first_code & 0xffffff) as usize + search_addr.unwrap()
                } else {
                    (first_code & 0xffffff) as usize
                };
                let mut count = (codes[pos + 2].1 & 0xfff0000) >> 16;
                let addr_diff = codes[pos + 2].1 & 0xffff;
                let val_diff = codes[pos + 3].1;
                while count != 0 {
                    if first_code & 0xf000000 == 0x1000000 {
                        if addr >= max_data_len {
                            return Err(format!("apply code error at line {}: address overflow", line));
                        }
                        data[addr] = (val & 0xff) as u8;
                    } else if first_code & 0xf000000 == 0x2000000 {
                        if addr + 1 >= max_data_len {
                            return Err(format!("apply code error at line {}: address overflow", line));
                        }
                        data[addr] = (val & 0xff) as u8;
                        data[addr + 1] = ((val >> 8) & 0xff) as u8;
                    } else {
                        if addr + 3 >= max_data_len {
                            return Err(format!("apply code error at line {}: address overflow", line));
                        }
                        data[addr] = (val & 0xff) as u8;
                        data[addr + 1] = ((val >> 8) & 0xff) as u8;
                        data[addr + 2] = ((val >> 16) & 0xff) as u8;
                        data[addr + 3] = ((val >> 24) & 0xff) as u8;
                    }
                    count -= 1;
                    addr += addr_diff as usize;
                    val += val_diff;
                }
            }
            0x5 => {
                return Err(format!("apply code error at line {}: not supported", line));
            }
            0x8 => {
                let count = (codes[pos].1 & 0xfff0000) >> 16;
                let search_bytes = (first_code & 0xffff) as usize;
                code_len = (search_bytes - 1) / 4 + 2;
                code_len += code_len % 2;
                let mut bytes = 0;
                let mut search = Vec::with_capacity(search_bytes);
                while bytes < search_bytes {
                    let offset = pos + 1 + (bytes / 4);
                    let shift = (3 - (bytes % 4)) * 8;
                    search.push(((codes[offset].1 >> shift) & 0xff) as u8);
                    bytes += 1;
                }
                search_addr =  search_data(count, &data, &search);
            }
            0xa => {
                let mut addr = (codes[pos].1 & 0xfffffff) as usize;
                let write_bytes = codes[pos + 1].1 as usize;
                code_len = ((write_bytes - 1) / 8 + 1) * 2 + 2;
                if addr + write_bytes>= max_data_len {
                    return Err(format!("apply code error at line {}: address overflow", line));
                }
                let mut bytes = 0;
                while bytes < write_bytes {
                    let offset = pos + 2 + (bytes / 4);
                    let shift = (3 - (bytes % 4)) * 8;
                    data[addr] = ((codes[offset].1 >> shift) & 0xff) as u8;
                    bytes += 1;
                    addr += 1;
                }                
            }
            _ => ()
        }
    }
    let patched_file = patch_file.to_string() + ".patched";
    let mut write_file = fs::OpenOptions::new().create(true).write(true).truncate(true).open(patched_file.as_str()).map_err(fn_map_err)?;
    write_file.write(data.as_slice()).map_err(fn_map_err)?;
    Ok(())
}

fn search_data(mut count : u32, data : &Vec<u8>, search : &Vec<u8>) -> Option<usize> {
    for (position, window) in data.windows(search.len()).enumerate() {
        if window == search {
            count -= 1;
            if count == 0 {
                return Some(position);
            }
        }
    }
    None
}
